////////////////////////////////////////////////////////////////////// 
// libsrc/MathNormal.h 
// (c) 2000-2001 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday December 14, 2001
// 
 
#ifndef __NORMALEQUATIONS_H__
#define __NORMALEQUATIONS_H__

#include "MathMatrix.h"
#include "MathVector.h"
#include "MathCholesky.h"

#define NORMAL_AMOEBA_MIN    0
#define NORMAL_POWELL_MIN    1

class NormalEquations
   {
   public:
      Vector   means, variances;

      Matrix * varComponents;
      Matrix   linearModel;
      Vector   scores;

      double   likelihood;

      NormalEquations();
      virtual ~NormalEquations();

      void   Dimension(int vcCount);

      virtual void   Prepare();
      virtual void   SelectPoint(Vector & v);
      virtual void   RetrievePoint(Vector & point);
      virtual double Evaluate();

      // Diagnostic statistics
      // see JL Hopper and JD Matthews Ann Hum Genet (1992) 46:373 - 383
      double rawQ;      // This is a chi-square with n degrees of freedom
      double Q;         // This is Q1 and has a standard normal distribution
      Vector Qi;        // Each Qi is approximately chi-square with 1 df
      void   Diagnostics();

      Matrix    varMatrix;
      Vector    residuals;
      Cholesky  cholesky;
      double    constant;
      bool      includeLikelihoodConstant, preliminary;
      int       multiple;

      bool operator == (const NormalEquations & rhs);

   protected:
      void Free();

      void CalculateResiduals();
      void CalculateCovariances();
   };

class NormalSet
   {
   public:
      NormalEquations ** sets;

      double precision;
      int    numericMinimizer;
      int    size;
      int    count;
      int    maxThreads;
      double likelihood;
      Vector variances, means;
      bool   iterative, preliminary;

      NormalSet(int threads = 0);

      virtual ~NormalSet() { Free(); }

      void        Dimension(int setCount, int vcCount);
      double      Evaluate();
      void        SelectPoint(Vector & v);
      void        Solve();
      int         CountObservations();
      virtual int CountParameters();

      NormalEquations & operator [] (int n)
         { return *(sets[n]); }

   protected:
      // for multi-threading
      static void * EvaluateOneSet(void * which);

      // house-keeping
      void  Free();
      virtual void AllocateSets();

      // Helpers for solver
      void         EditLinearDegenerates();
      virtual void GetStartingPoint(Vector & startPoint);
      void         RemoveRedundancy();
   };

class NormalSolver : public VectorFunc
   {
   public:
      NormalSet * normal;

      NormalSolver(NormalSet * n) : VectorFunc()
         { normal = n; }

      virtual double Evaluate(Vector & point);
   };

#endif


 
