////////////////////////////////////////////////////////////////////// 
// merlin/VarianceComponents.cpp 
// (c) 2000-2001 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday December 14, 2001
// 
 
#include "VarianceComponents.h"

#include "MathStats.h"
#include "MathNormal.h"

#include <math.h>

void VarianceComponents::Analyse(Pedigree & ped, MerlinKinship & markerKin, StringArray & labels)
   {
   int positions = labels.Length();

   IntArray * pheno = new IntArray[ped.familyCount];
   Kinship kin;

   // Loop through traits in the pedigree
   for (int t = 0; t < ped.traitCount; t++)
      {
      // First list phenotyped individuals for each family
      for (int f = 0; f < ped.familyCount; f++)
         {
         pheno[f].Dimension(0);
         if (markerKin.CheckFamily(f))
            for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
               if (ped[i].isPhenotyped(t))
                  pheno[f].Push(i);
         }

      // Number of families with non-null kinships and phenotypes
      int useful = 0;

      // Count useful families
      for (int f = 0; f < ped.familyCount; f++)
         if (pheno[f].Length())
            useful++;

      if (useful == 0)
         {
         printf("Trait: %s (No informative families)\n\n",
                (const char *) ped.traitNames[t]);
         continue;
         }

      // The normal set class is our workhorse
      NormalSet null;
      null.Dimension(useful, 2);

      // Index to the normal equations within NormalSet
      int index = 0;

      // Build model with polygenes and environment only
      for (int f = 0; f < ped.familyCount; f++)
         if (pheno[f].Length())
            {
            int count = pheno[f].Length();

            // Phenotypes
            null[index].scores.Dimension(count);
            for (int i = 0; i < count; i++)
               null[index].scores[i] = ped[pheno[f][i]].traits[t];

            // Variable for regressing grand mean
            null[index].linearModel.Dimension(count, 1);
            null[index].linearModel.Set(1.0);

            // Setup non-shared variances
            null[index].varComponents[0].Dimension(count, count);
            null[index].varComponents[0].Identity();

            // Setup polygenic variances
            null[index].varComponents[1].Dimension(count, count);
            kin.Setup(*ped.families[f]);

            for (int i = 0; i < count; i++)
               for (int j = i; j < count; j++)
                  null[index].varComponents[1][i][j] =
                  null[index].varComponents[1][j][i] =
                     2.0 * kin(ped[pheno[f][i]], ped[pheno[f][j]]);

            index++;
            }

      // Fit polygenic model
      null.Solve();

      // Track Key Values
      double lkNull = null.Evaluate();
      double sampleVar = null.variances.Sum();
      double sampleH2 = null.variances[1];

      // Print out header for results
      printf("Phenotype: %s [VC] (%d famil%s, h2 = %.2f%%)\n",
            (const char *) ped.traitNames[t], useful, useful == 1 ? "y" : "ies",
            sampleH2 * 100. / sampleVar);
      printf("=====================================================\n");
      printf("%20s %8s %7s %7s %7s\n",
             "Position", "H2 ", "ChiSq", "LOD", "pvalue");

      // The normal set class is our workhorse
      NormalSet alternate;
      alternate.Dimension(useful, 3);

      // Build model with polygenes and environment only
      for (int f = 0; f < useful; f++)
         {
         int count = null[f].linearModel.rows;
         alternate[f].scores = null[f].scores;
         alternate[f].linearModel = null[f].linearModel;
         alternate[f].varComponents[0] = null[f].varComponents[0];
         alternate[f].varComponents[1] = null[f].varComponents[1];
         alternate[f].varComponents[2].Dimension(count, count);
         }

      // Loop through analysis at individual marker locations
      for (int pos = 0; pos < positions; pos++)
         {
         index = 0;

         // Fill major gene matrix with observed coefficients
         for (int f = 0; f < ped.familyCount; f++)
            if (pheno[f].Length())
               {
               int count = alternate[index].linearModel.rows;

               for (int i = 0; i < count; i++)
                  for (int j = i; j < count; j++)
                     alternate[index].varComponents[2][i][j] =
                     alternate[index].varComponents[2][j][i] =
                                  2.0 * markerKin.Retrieve(f, pos,
                                        ped[pheno[f][i]].traverse,
                                        ped[pheno[f][j]].traverse);

               // Labels for kinship matrices, useful for debugging
               // but usually not necessary
               // alternate[index].varComponents[2].SetLabel(ped[pheno[f][0]].famid);
               // for (int j = 0; j < count; j++)
               //   {
               //   alternate[index].varComponents[2][j].label = ped[pheno[f][j]].pid;
               //   alternate[index].varComponents[2].SetColumnLabel(j, ped[pheno[f][j]].pid);
               //   }

               index++;
               }

         index = 0;

         // Fit the alternative model
         alternate.Solve();

         // Evaluate the likelihood
         double lk = alternate.Evaluate();

         // Get key values
         double chisq = lkNull > lk ? (lkNull - lk) * 2.0 : 0.0;
         double pvalue = chidist(chisq, 1) * 0.5;
         double h2 = alternate.variances[2];
         double var = alternate.variances.Sum();

         int digits  = pvalue < 1.5e-4 ? 5 : int(log(pvalue*0.06666666)*-0.4343);

         // Print out a one-line summary of results
         printf("%20.20s %7.2f%% %7.2f %7.2f %7.*f\n",
                (const char *) labels[pos],
                h2 * 100 / var, chisq, chisq / (2*log(10)), digits, pvalue);
         }
      }

   delete [] pheno;
   }

 
